/*
Copyright (c) 2020 Krzysztof Mitko

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdio.h>
#include <stdlib.h>
#include "debiensko.h"
#include "nelder_mead.h"

brine_t lyes_tmp;

static void rcc_fun(int n, point_t* point, const void* arg) {
  const brine_t *feed = (const brine_t*)arg;
  double mCl, mCa, mMg, mSO, mNa, mH2O, d;
  mCl = feed->cl * feed->v;
  mCa = feed->ca * feed->v;
  mMg = feed->mg * feed->v;
  mSO = feed->so * feed->v;
  mNa = Na(*feed) * feed->v;
  mH2O = density(*feed)*feed->v - mCl - mCa - mMg - mSO - mNa;
  mCl = mCl - point->x[0]/58.5*35.5;
  mCa = mCa - point->x[1]/172*40;
  mSO = mSO - point->x[1]/172*96;
  mH2O = mH2O - point->x[1]/172*36 - point->x[2];
  mNa = mNa - point->x[0]/58.5*23;
  d = 7.7526*((mCl/35.5*58.5)/(density(*feed)*feed->v))*100+997.05;
  lyes_tmp.v = (mCl + mCa + mSO + mMg + mNa + mH2O)/d;
  lyes_tmp.cl = mCl/lyes_tmp.v;
  lyes_tmp.ca = mCa/lyes_tmp.v;
  lyes_tmp.mg = mMg/lyes_tmp.v;
  lyes_tmp.so = mSO/lyes_tmp.v;
  point->fx = pow(MAXNACL - lyes_tmp.cl,2)/(MAXNACL*MAXNACL) + pow(MAXBIVALENTS - 111*mCa/(lyes_tmp.v*d*40) - 95*mMg/(lyes_tmp.v*d*24),2)/(MAXBIVALENTS*MAXBIVALENTS) + pow(MAXGYPSUM - pow(lyes_tmp.ca/40,2)*pow(lyes_tmp.so/96,2),2)/(MAXGYPSUM*MAXGYPSUM);
  //if (point->x[0] < 0 || point->x[1] < 0 || point->x[2] < 0 || lyes_tmp.cl < 0 || lyes_tmp.ca < 0 || lyes_tmp.mg < 0 || lyes_tmp.so < 0 || lyes_tmp.v <0) point->fx = point->fx * 1e6;
}

double Na(brine_t brine) {
  return 23*(brine.cl/35.5 + brine.so/96*2 - brine.ca/40*2 - brine.mg/24*2);
}
double TDSmass(brine_t brine) {
  return(brine.cl + brine.ca + brine.mg + brine.so + Na(brine));
}
double density(brine_t brine) {
  //double T = 298;
  double sal = brine.cl/35.5*58.5;
  return 9.248e-10*pow(sal,4)-2.834e-7*pow(sal,3)-2.26518e-4*pow(sal,2)+0.6798978*sal+1004.0208;
  //return 9.248e-10*pow(sal,4)-2.834e-7*pow(sal,3)+(-1.41e-7*T-1.845e-4)*pow(sal,2)+(2.161e-4*T+0.6155)*sal+5.375e-5*pow(T,2)+0.3062*T+908;
}
double TDS(brine_t brine) {
  return 100 * TDSmass(brine) / density(brine);
}
brine_t ConcentrateVC(brine_t feed) {
  brine_t brine;
  brine.cl = CHLORIDELIMIT;
  brine.v = feed.cl / brine.cl * feed.v;
  brine.ca = feed.ca * feed.v / brine.v;
  brine.mg = feed.mg * feed.v / brine.v;
  brine.so = feed.so * feed.v / brine.v;
  return brine;
}
double DiluateVC(brine_t feed) {
  return (feed.v - feed.cl / CHLORIDELIMIT * feed.v);
}
double EnergyVC() {
  return 44.0;
}
double EnergyRCC() {
  return 66.0;
}
brine_t DiluateRO(brine_t feed, double recovery) {
  brine_t permeate;
  permeate.cl = 0.0078*feed.cl;
  permeate.ca = 0.0015*feed.ca;
  permeate.mg = 0.001*feed.mg;
  permeate.so = 0.0028*feed.so;
  permeate.v = recovery*feed.v/100;
  return permeate;
}
brine_t ConcentrateRO(brine_t feed, brine_t permeate) {
  brine_t retentate;
  retentate.v = feed.v - permeate.v;
  retentate.cl = (feed.cl*feed.v - permeate.cl*permeate.v)/retentate.v;
  retentate.ca = (feed.ca*feed.v - permeate.ca*permeate.v)/retentate.v;
  retentate.mg = (feed.mg*feed.v - permeate.mg*permeate.v)/retentate.v;
  retentate.so = (feed.so*feed.v - permeate.so*permeate.v)/retentate.v;
  return retentate;
}
double Pf(brine_t feed, brine_t retentate) {
  return (0.448175 * (TDSmass(feed) + TDSmass(retentate)) + 24);
}
double EnergyRO(double recovery, double pressure, int hasEnergyRecovery) {
  double result;
  result = 0.02724*pressure/(recovery/100*0.88*0.96) + 0.05;
  if (hasEnergyRecovery != 0) {
    result = result - 0.02724*(1-recovery/100)/(recovery/100)*0.94/0.96*(pressure-2);
  }
  return result;
}
rcc_results_t ConcentrateRCC(brine_t feed) {
  optimset_t optimset;
  point_t solution;
  point_t start;
  rcc_results_t postcryst;
  postcryst.gypsum = 0;
  start.x = malloc(3 * sizeof(double));
  start.x[0] = 0.4*feed.cl*feed.v;
  start.x[1] = 0;
  start.x[2] = 0.4*feed.v*density(feed);
  optimset.tolx = 0.000001;
  optimset.tolf = 0.000001;
  optimset.max_iter = 1000;
  optimset.max_eval = 1000;
  optimset.verbose = 0;
  nelder_mead(3, &start, &solution, &rcc_fun, &feed, &optimset);
  postcryst.lyes = lyes_tmp;
  postcryst.salt = solution.x[0];
  postcryst.gypsum = solution.x[1];
  postcryst.water = solution.x[2];
  free(start.x);
  free(solution.x);
  return(postcryst);
}

void BrinePrintf(brine_t brine) {
  printf("Chlorides : %#12.4g kg/m3\n", brine.cl);
  printf("Calcium   : %#12.4g kg/m3\n", brine.ca);
  printf("Magnesium : %#12.4g kg/m3\n", brine.mg);
  printf("Sulphates : %#12.4g kg/m3\n", brine.so);
  printf("Sodium    : %#12.4g kg/m3\n", Na(brine));
  printf("TDS       : %#12.4g %%\n", TDS(brine));
  printf("Density   : %12.4g kg/m3\n", density(brine));
  printf("Flow      : %#12.4g m3/h\n", brine.v);
}

/* void ExcelBrinePrintf(FILE* fd, brine_t brine) {
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Chlorides [kg/m3]");
  ExcelPrintNumberCell(fd, brine.cl);
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Calcium [kg/m3]");
  ExcelPrintNumberCell(fd, brine.ca);
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Magnesium [kg/m3]");
  ExcelPrintNumberCell(fd, brine.mg);
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Sulphates [kg/m3]");
  ExcelPrintNumberCell(fd, brine.so);
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Sodium [kg/m3]");
  ExcelPrintNumberCell(fd, Na(brine));
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "TDS [kg/m3]");
  ExcelPrintNumberCell(fd, TDS(brine));
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Density [kg/m3]");
  ExcelPrintNumberCell(fd, density(brine));
  ExcelFinishRow(fd);
  ExcelStartNewRow(fd);
  ExcelPrintStringCell(fd, "Flow [m3/h]");
  ExcelPrintNumberCell(fd, brine.v);
  ExcelFinishRow(fd);
}

void ExcelPrintStringCell(FILE* fd, const char* content) {
  fprintf(fd, "<Cell><Data ss:Type=\"String\">%s</Data></Cell>\n", content);
}

void ExcelPrintNumberCell(FILE* fd, double content) {
  fprintf(fd, "<Cell><Data ss:Type=\"Number\">%g</Data></Cell>\n", content);
}

void ExcelPrintHeader(FILE* fd) {
  fprintf(fd, "<?xml version=\"1.0\"?>\n<Workbook xmlns=\"urn:schemas-microsoft-com:office:spreadsheet\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:x=\"urn:schemas-microsoft-com:office:excel\" xmlns:ss=\"urn:schemas-microsoft-com:office:spreadsheet\" xmlns:html=\"http://www.w3.org/TR/REC-html40\">\n<Worksheet ss:Name=\"Sheet1\">\n<Table>\n");
}

void ExcelStartNewRow(FILE* fd) {
  fprintf(fd, "<Row>\n");
}

void ExcelFinishRow(FILE* fd) {
  fprintf(fd, "</Row>\n");
}

void ExcelPrintFinish(FILE* fd) {
  fprintf(fd, "</Table></Worksheet></Workbook>");
}
*/
