/*
Copyright (c) 2020 Krzysztof Mitko

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#ifndef debiensko_h
#define debiensko_h

#include <stdio.h>
#include <math.h>

#define EXPORT __attribute__((visibility("default")))

#define CHLORIDELIMIT 176.0
#define MAXNACL 200.0
#define MAXBIVALENTS 0.08
#define MAXGYPSUM 4.30165423622131E-06

// Data type brine_t describes composition and flow rate of a process stream
typedef struct {
  double cl; // Chlorides [kg/m3]
  double ca; // Calcium [kg/m3]
  double mg; // Magnesium [kg/m3]
  double so; // Sulphates [kg/m3]
  double v; // Flow rate [m3/h]
} brine_t;

// Data type rcc_results_t describes the amount of removed salt, water, and gypsum [kg/h] by the crystallizer, as well as composition of post-crystallization lyes [brine_t]
typedef struct {
  double salt;
  double water;
  double gypsum;
  brine_t lyes;
} rcc_results_t;

// Calculate amount of sodium [kg/m3] for a given process stream [brine_t]
EXPORT double Na(brine_t brine);
// Calculate TDS [kg/m3] of a process stream [brine_t]
EXPORT double TDSmass(brine_t brine);
// Calculate density [kg/m3] of a process stream [brine_t], assumed temperature 298 K
EXPORT double density(brine_t brine);
// Calculate TDS [%] of a process stream [brine_t]
EXPORT double TDS(brine_t brine);
// Calculate the evaporator concentrate [brine_t] based on the feed [brine_t]
EXPORT brine_t ConcentrateVC(brine_t feed);
// Calculate the volume of eaporator distillate [m3/h] based on the feed [brine_t]
EXPORT double DiluateVC(brine_t feed);
// Calculate the energy consumption of evaporator [kWh/m3 of distillate] based on feed stream [brine_t]
// This is only a placeholder - for Dębieńsko it's always 44 kWh/m3 of distillate
EXPORT double EnergyVC(void);
// Calculate the RO permeate [brine_t] at given recovery [%]
EXPORT brine_t DiluateRO(brine_t feed, double recovery);
// Calculate RO retentate [brine_t] based on mass balance of feed [brine_t] and permeate [brine_t]
EXPORT brine_t ConcentrateRO(brine_t feed, brine_t permeate);
// Calculate the pressure [bar] required by RO given the feed [brine_t] and retentate [brine_t]
EXPORT double Pf(brine_t feed, brine_t retentate);
// Calculate the RO energy consumption [kWh/m3 of permeate] at given pressure [bar] and recovery [%]
// If hasEnergyRecovery is set to 0, then no ERD is assumed
EXPORT double EnergyRO(double recovery, double pressure, int hasEnergyRecovery);
// Calculate the energy consumption of crystallizer [kWh/m3 of distillate] based on feed stream [brine_t]
// This is only a placeholder - for Dębieńsko it's always 66 kWh/m3 of distillate
EXPORT double EnergyRCC(void);
// Calculate the post-crystallization lyes and the amount of removed salt, water, gypsum
EXPORT rcc_results_t ConcentrateRCC(brine_t feed);

// Printout the brine parameters to stdout
EXPORT void BrinePrintf(brine_t brine);

/* //Export results to xml
EXPORT void ExcelPrintStringCell(FILE* fd, const char* content);
EXPORT void ExcelPrintNumberCell(FILE* fd, double content);
EXPORT void ExcelPrintHeader(FILE* fd);
EXPORT void ExcelPrintFinish(FILE* fd);
EXPORT void ExcelStartNewRow(FILE* fd);
EXPORT void ExcelFinishRow(FILE* fd);
EXPORT void ExcelBrinePrintf(FILE* fd, brine_t brine);
*/

#endif /* debiensko_h */
